// SPDX-License-Identifier: LGPL-2.1
// Copyright 2021 Daniel Vogelbacher <daniel@chaospixel.com>

use image::DynamicImage;
use log::{debug, warn};
use std::convert::{TryFrom, TryInto};
use std::f32::NAN;
use std::fmt::Debug;

use crate::bits::Endian;
use crate::decompressors::crx::decompress_crx_image;
use crate::envparams::{rawler_crx_raw_trak, rawler_ignore_previews};
use crate::formats::bmff::ext_cr3::cmp1::Cmp1Box;
use crate::formats::bmff::ext_cr3::cr3desc::Cr3DescBox;
use crate::formats::bmff::ext_cr3::iad1::{Iad1Box, Iad1Type};
use crate::formats::bmff::trak::TrakBox;
use crate::formats::bmff::FileBox;
use crate::formats::tiff::reader::TiffReader;
use crate::formats::tiff::{Entry, GenericTiffReader, Rational, Value};
use crate::imgop::{Point, Rect};
use crate::lens::{LensDescription, LensResolver};
use crate::tags::{DngTag, TiffTagEnum};
use crate::{decoders::*, RawFile};
use crate::{pumps::ByteStream, RawImage};

/// Decoder for CR3 and CRM files
#[allow(dead_code)]
#[derive(Debug, Clone)]
pub struct Cr3Decoder<'a> {
  camera: Camera,
  rawloader: &'a RawLoader,
  bmff: Bmff,
  exif: Option<GenericTiffReader>,
  makernotes: Option<GenericTiffReader>,
  wb: Option<[f32; 4]>,
  blacklevels: Option<[u16; 4]>,
  whitelevel: Option<u16>,
  // Basic EXIF information
  cmt1: GenericTiffReader,
  // EXIF
  cmt2: GenericTiffReader,
  // Makernotes
  cmt3: GenericTiffReader,
  // GPS
  cmt4: GenericTiffReader,
  ctmd_exposure: Option<CtmdExposureInfo>,
  ctmd_focallen: Option<Rational>,
  ctmd_rec7_exif: Option<GenericTiffReader>,
  ctmd_rec7_makernotes: Option<GenericTiffReader>,
  // CTMD Makernotes: COLORDATA
  ctmd_rec8: Option<GenericTiffReader>,
  ctmd_rec9: Option<GenericTiffReader>,
  xpacket: Option<Vec<u8>>,
  image_unique_id: Option<[u8; 16]>,
  lens_description: Option<&'static LensDescription>,
}

#[allow(dead_code)]
const CR3_CTMD_BLOCK_EXIFIFD: u16 = 0x8769;
const CR3_CTMD_BLOCK_MAKERNOTES: u16 = 0x927c;

/// Type values fro CCTP records
#[derive(Clone, Copy, Debug)]
#[allow(dead_code)]
enum Cr3ImageType {
  CrxBix = 0,
  CrxSmall = 1,
  PreviewBig = 2,
  Ctmd = 3,
  CrxDual = 4,
}

impl<'a> Cr3Decoder<'a> {
  /// Construct new CR3 or CRM deocder
  pub fn new(_rawfile: &mut RawFile, bmff: Bmff, rawloader: &'a RawLoader) -> Result<Cr3Decoder<'a>> {
    if let Some(Cr3DescBox { cmt1, cmt2, cmt3, cmt4, .. }) = bmff.filebox.moov.cr3desc.as_ref() {
      let _mode = Self::get_mode(cmt3.tiff.root_ifd())?;
      let mode = ""; // TODO add modes
      let camera = rawloader.check_supported_with_mode(cmt1.tiff.root_ifd(), mode)?;
      let decoder = Cr3Decoder {
        camera,
        rawloader,
        exif: None,
        makernotes: None,
        wb: None,
        blacklevels: None,
        whitelevel: None,
        cmt1: cmt1.tiff.clone(),
        cmt2: cmt2.tiff.clone(),
        cmt3: cmt3.tiff.clone(),
        cmt4: cmt4.tiff.clone(),
        xpacket: None,
        image_unique_id: None,
        lens_description: None,
        ctmd_rec7_exif: None,
        ctmd_rec7_makernotes: None,
        ctmd_rec8: None,
        ctmd_rec9: None,
        ctmd_exposure: None,
        bmff,
        ctmd_focallen: None,
      };
      return Ok(decoder);
    } else {
      return Err(RawlerError::Unsupported(format!("It's not a CR3 file: CMT boxes not found.")));
    }
  }

  // Search for quality tag inside makernotes and derive
  // our mode string for configuration.
  fn get_mode(makernotes: &IFD) -> Result<&str> {
    Ok(if let Some(entry) = makernotes.get_entry(0x0001) {
      match entry.get_u16(3)? {
        Some(4) => "raw",
        Some(7) => "craw",
        Some(130) => "crm",
        Some(131) => "crm",
        _ => "undefined",
      }
    } else {
      "undefined"
    })
  }

  /// Get trak from moov box
  fn moov_trak(&self, trak_id: usize) -> Option<&TrakBox> {
    self.bmff.filebox.moov.traks.get(trak_id)
  }

  /// Get IAD1 box for specific trak
  fn iad1_box(&self, trak_idx: usize) -> Option<&Iad1Box> {
    let trak = &self.bmff.filebox.moov.traks[trak_idx];
    let craw = trak.mdia.minf.stbl.stsd.craw.as_ref();
    craw.and_then(|craw| craw.cdi1.as_ref()).and_then(|cdi1| Some(&cdi1.iad1))
  }

  /// Get CMP1 box for specific trak
  fn cmp1_box(&self, trak_idx: usize) -> Option<&Cmp1Box> {
    let trak = &self.bmff.filebox.moov.traks[trak_idx];
    let craw = trak.mdia.minf.stbl.stsd.craw.as_ref();
    craw.and_then(|craw| craw.cmp1.as_ref())
  }

  /// Read CTMD records for given sample
  /// Each sample (for movie files) have their own CTMD records
  fn read_ctmd(&self, rawfile: &mut RawFile, sample_idx: u32) -> Result<Option<Ctmd>> {
    let ctmd_trak = &self.bmff.filebox.moov.traks[3];
    let (offset, size) = ctmd_trak.mdia.minf.stbl.get_sample_offset(sample_idx as u32 + 1).unwrap();
    debug!("CR3 CTMD mdat offset for sample_idx {}: {}, len: {}", sample_idx, offset, size);
    let buf = rawfile
      .get_range(offset as u64, size as u64)
      .map_err(|e| RawlerError::General(format!("I/O error while reading CR3 CTMD: {:?}", e)))?;
    let mut substream = ByteStream::new(&buf, Endian::Little);
    let ctmd = Ctmd::new(&mut substream);
    Ok(Some(ctmd))
  }
}

const EXIF_TRANSFER_TAGS: [u16; 23] = [
  ExifTag::ExposureTime as u16,
  ExifTag::FNumber as u16,
  ExifTag::ISOSpeedRatings as u16,
  ExifTag::SensitivityType as u16,
  ExifTag::RecommendedExposureIndex as u16,
  ExifTag::ISOSpeed as u16,
  ExifTag::FocalLength as u16,
  ExifTag::ExposureBiasValue as u16,
  ExifTag::DateTimeOriginal as u16,
  ExifTag::CreateDate as u16,
  ExifTag::OffsetTime as u16,
  ExifTag::OffsetTimeDigitized as u16,
  ExifTag::OffsetTimeOriginal as u16,
  ExifTag::OwnerName as u16,
  ExifTag::LensSerialNumber as u16,
  ExifTag::SerialNumber as u16,
  ExifTag::ExposureProgram as u16,
  ExifTag::MeteringMode as u16,
  ExifTag::Flash as u16,
  ExifTag::ExposureMode as u16,
  ExifTag::WhiteBalance as u16,
  ExifTag::SceneCaptureType as u16,
  ExifTag::ShutterSpeedValue as u16,
];

fn transfer_exif_tag(tag: u16) -> bool {
  EXIF_TRANSFER_TAGS.contains(&tag)
}

#[derive(Debug, Clone, PartialEq, Default, Serialize, Deserialize)]
#[serde(rename_all = "camelCase")]
pub struct Cr3Format {
  pub filebox: FileBox,
  pub ctmd_exposure: Option<CtmdExposureInfo>,
  pub ctmd_rec7_exif: Option<GenericTiffReader>,
  pub ctmd_rec7_makernotes: Option<GenericTiffReader>,
  pub ctmd_rec8: Option<GenericTiffReader>,
  pub ctmd_rec9: Option<GenericTiffReader>,
}

impl<'a> Decoder for Cr3Decoder<'a> {
  fn format_dump(&self) -> FormatDump {
    FormatDump::Cr3(Cr3Format {
      filebox: self.bmff.filebox.clone(),
      ctmd_exposure: self.ctmd_exposure.clone(),
      ctmd_rec7_exif: self.ctmd_rec7_exif.clone(),
      ctmd_rec7_makernotes: self.ctmd_rec7_makernotes.clone(),
      ctmd_rec8: self.ctmd_rec8.clone(),
      ctmd_rec9: self.ctmd_rec9.clone(),
    })
  }

  fn xpacket(&self, _file: &mut RawFile) -> Option<&Vec<u8>> {
    self.xpacket.as_ref()
  }

  fn populate_capture_info(&mut self, capture_info: &mut CaptureInfo) -> Result<()> {
    let ifd = self.cmt2.root_ifd();
    if let Some(Entry { value: Value::Rational(v), .. }) = ifd.get_entry(ExifTag::ExposureTime) {
      capture_info.exposure_time = Some(v[0])
    }
    if let Some(Entry {
      value: Value::SRational(v), ..
    }) = ifd.get_entry(ExifTag::ExposureBiasValue)
    {
      capture_info.exposure_bias = Some(v[0])
    }
    if let Some(Entry {
      value: Value::SRational(v), ..
    }) = ifd.get_entry(ExifTag::ShutterSpeedValue)
    {
      capture_info.shutter_speed = Some(v[0])
    }

    if let Some(lens) = self.lens_description {
      let lens_spec: [Rational; 4] = [lens.focal_range[0], lens.focal_range[1], lens.aperture_range[0], lens.aperture_range[1]];
      capture_info.lens_make = Some(lens.lens_make.clone());
      capture_info.lens_model = Some(lens.lens_model.clone());
      capture_info.lens_spec = Some(lens_spec);
    }

    Ok(())
  }

  fn populate_dng_root(&mut self, root_ifd: &mut DirectoryWriter) -> Result<()> {
    // Copy Orientation tag

    let ifd = self.cmt1.root_ifd();
    if let Some(orientation) = ifd.get_entry(ExifTag::Orientation) {
      root_ifd.add_value(ExifTag::Orientation, orientation.value.clone())?;
    }
    if let Some(artist) = ifd.get_entry(ExifTag::Artist) {
      root_ifd.add_value(ExifTag::Artist, artist.value.clone())?;
    }
    if let Some(copyright) = ifd.get_entry(ExifTag::Copyright) {
      root_ifd.add_value(ExifTag::Copyright, copyright.value.clone())?;
    }

    if let Some(lens) = self.lens_description {
      let lens_info: [Rational; 4] = [lens.focal_range[0], lens.focal_range[1], lens.aperture_range[0], lens.aperture_range[1]];
      root_ifd.add_tag(DngTag::LensInfo, lens_info)?;
    }

    if let Some(unique_id) = self.image_unique_id {
      // For CR3, we use the already included Makernote tag with unique image ID
      root_ifd.add_tag(DngTag::RawDataUniqueID, unique_id)?;
    }

    let gpsinfo_offset = {
      let mut gps_ifd = root_ifd.new_directory();
      let ifd = self.cmt4.root_ifd();
      // Copy all GPS tags
      for (tag, entry) in ifd.entries() {
        match tag {
          // Special handling for Exif.GPSInfo.GPSLatitude and Exif.GPSInfo.GPSLongitude.
          // Exif.GPSInfo.GPSTimeStamp is wrong, too and can be fixed with the same logic.
          // Canon CR3 contains only two rationals, but these tags are specified as a vector
          // of three reationals (degrees, minutes, seconds).
          // We fix this by extending with 0/1 as seconds value.
          0x0002 | 0x0004 | 0x0007 => match &entry.value {
            Value::Rational(v) => {
              let fixed_value = if v.len() == 2 { vec![v[0], v[1], Rational::new(0, 1)] } else { v.clone() };
              gps_ifd.add_value(*tag, Value::Rational(fixed_value))?;
            }
            _ => {
              warn!("CR3: Exif.GPSInfo.GPSLatitude and Exif.GPSInfo.GPSLongitude expected to be of type RATIONAL, GPS data is ignored");
            }
          },
          _ => {
            gps_ifd.add_value(*tag, entry.value.clone())?;
          }
        }
      }
      gps_ifd.build()?
    };
    root_ifd.add_tag(ExifTag::GPSInfo, gpsinfo_offset as u32)?;

    Ok(())
  }

  fn populate_dng_exif(&mut self, exif_ifd: &mut DirectoryWriter) -> Result<()> {
    let ifd = self.cmt2.root_ifd();
    for (tag, entry) in ifd.entries().iter().filter(|(tag, _)| transfer_exif_tag(**tag)) {
      exif_ifd.add_value(*tag, entry.value.clone())?;
    }

    if let Some(focal_len) = self.ctmd_focallen {
      exif_ifd.add_tag(ExifTag::FocalLength, focal_len)?;
    }

    if let Some(lens) = self.lens_description {
      let lens_info: [Rational; 4] = [lens.focal_range[0], lens.focal_range[1], lens.aperture_range[0], lens.aperture_range[1]];
      exif_ifd.add_tag(ExifTag::LensSpecification, lens_info)?;
      exif_ifd.add_tag(ExifTag::LensMake, &lens.lens_make)?;
      exif_ifd.add_tag(ExifTag::LensModel, &lens.lens_model)?;
    }

    Ok(())
  }

  /// CR3 can store multiple samples in trak
  fn raw_image_count(&self) -> Result<usize> {
    let raw_trak_id = rawler_crx_raw_trak()
      .or(self.get_trak_index(Cr3ImageType::CrxBix))
      .ok_or(RawlerError::General("Unable to find trak index".into()))?;
    let moov_trak = self.moov_trak(raw_trak_id).ok_or(format!("Unable to get MOOV trak {}", raw_trak_id))?;
    Ok(moov_trak.mdia.minf.stbl.stsz.sample_count as usize)
  }

  /// Decode raw image
  fn raw_image(&mut self, file: &mut RawFile, params: RawDecodeParams, dummy: bool) -> Result<RawImage> {
    let raw_trak_id = rawler_crx_raw_trak()
      .or(self.get_trak_index(Cr3ImageType::CrxBix))
      .ok_or(RawlerError::General("Unable to find trak index".into()))?;
    let sample_idx = params.image_index;
    if sample_idx >= self.raw_image_count()? {
      return Err(RawlerError::General(format!(
        "Raw image index {} out of range ({})",
        sample_idx,
        self.raw_image_count()?
      )));
    }
    self.parse_makernotes(file)?;
    self.load_ctmd(file, sample_idx as u32)?;
    // Load trak with raw MDAT section
    let moov_trak = self.moov_trak(raw_trak_id).ok_or(format!("Unable to get MOOV trak {}", raw_trak_id))?;
    let (offset, size) = moov_trak.mdia.minf.stbl.get_sample_offset(sample_idx as u32 + 1).unwrap();
    debug!("RAW mdat offset: {}, len: {}", offset, size);
    // Raw data buffer
    let buf = file
      .get_range(offset as u64, size as u64)
      .map_err(|e| RawlerError::General(format!("I/O error while reading CR3 raw image: {:?}", e)))?;

    let cmp1 = self
      .cmp1_box(raw_trak_id)
      .ok_or(RawlerError::General(format!("CMP1 box not found for trak {}", raw_trak_id)))?;
    debug!("cmp1 mdat hdr size: {}", cmp1.mdat_hdr_size);

    let mut wb = self.wb.unwrap_or([NAN, NAN, NAN, NAN]);
    let whitelevel = self.whitelevel.unwrap_or(u16::MAX);

    // Special handling for CRM movie files
    if let Some(entry) = self.cmt3.get_entry(0x0001) {
      if Some(130) == entry.get_u16(3)? {
        // Light Raw
        // WB is already applied, use 1.0
        wb = [1024.0, 1024.0, 1024.0, 1024.0];
      }
      if Some(131) == entry.get_u16(3)? { // Standard Raw
         // Nothing special for Standard raw
      }
    }

    let image = if !dummy {
      decompress_crx_image(&buf, cmp1).map_err(|e| format!("Failed to decode raw: {}", e.to_string()))?
    } else {
      Vec::new()
    };

    let cpp = 1;
    let mut img = RawImage::new(self.camera.clone(), cmp1.f_width as usize, cmp1.f_height as usize, cpp, wb, image, dummy);
    img.blacklevels = self.blacklevels.unwrap_or([0, 0, 0, 0]);
    img.whitelevels = [whitelevel, whitelevel, whitelevel, whitelevel];

    // IAD1 box contains sensor information
    // We use the sensor crop from IAD1 as recommended image crop.
    // The same crop is used as ActiveArea, because black areas in IAD1 are not
    // correct (they differs like 4-6 pixels from real values).
    match self.iad1_box(raw_trak_id) {
      Some(iad1) => {
        match &iad1.iad1_type {
          // IAD1 (small, used for CRM movie files)
          Iad1Type::Small(small) => {
            img.crop_area = Some(Rect::new_with_points(
              Point::new(small.crop_left_offset as usize, small.crop_top_offset as usize),
              Point::new((small.crop_right_offset + 1) as usize, (small.crop_bottom_offset + 1) as usize),
            ));
            img.active_area = img.crop_area;
          }
          // IAD1 (big, used for full size raws)
          Iad1Type::Big(big) => {
            img.crop_area = Some(Rect::new_with_points(
              Point::new(big.crop_left_offset as usize, big.crop_top_offset as usize),
              Point::new((big.crop_right_offset + 1) as usize, (big.crop_bottom_offset + 1) as usize),
            ));

            // Won't work!
            let _rect = Rect::new_with_points(
              Point::new(big.active_area_left_offset as usize, big.active_area_top_offset as usize),
              Point::new((big.active_area_right_offset - 1) as usize, (big.active_area_bottom_offset - 1) as usize),
            );
            //img.active_area = Some(rect);
            img.active_area = img.crop_area;

            let blackarea_h = Rect::new_with_points(
              Point::new(big.lob_left_offset as usize, big.lob_top_offset as usize),
              Point::new((big.lob_right_offset - 1) as usize, (big.lob_bottom_offset - 1) as usize),
            );
            if !blackarea_h.is_empty() {
              //img.blackareas.push(blackarea_h);
            }
            let blackarea_v = Rect::new_with_points(
              Point::new(big.tob_left_offset as usize, big.tob_top_offset as usize),
              Point::new((big.tob_right_offset - 1) as usize, (big.tob_bottom_offset - 1) as usize),
            );
            if !blackarea_v.is_empty() {
              //img.blackareas.push(blackarea_v);
            }
          }
        }
      }

      None => {
        warn!("No IAD1 box found for sensor data");
      }
    }
    debug!("Canon active area: {:?}", img.active_area);
    debug!("Canon crop area: {:?}", img.crop_area);
    debug!("Black areas: {:?}", img.blackareas);
    return Ok(img);
  }

  /// Extract preview image embedded in CR3
  fn full_image(&self, file: &mut RawFile) -> Result<DynamicImage> {
    if rawler_ignore_previews() {
      return Err(RawlerError::General("Unable to extract preview image".into()));
    }
    let offset = self.bmff.filebox.moov.traks[0].mdia.minf.stbl.co64.as_ref().expect("co64 box").entries[0] as usize;
    let size = self.bmff.filebox.moov.traks[0].mdia.minf.stbl.stsz.sample_sizes[0] as usize;
    debug!("JPEG preview mdat offset: {}, len: {}", offset, size);
    let buf = file
      .get_range(offset as u64, size as u64)
      .map_err(|e| RawlerError::General(format!("I/O error while reading CR3 full image: {:?}", e)))?;
    match image::load_from_memory_with_format(&buf, image::ImageFormat::Jpeg) {
      Ok(img) => Ok(img),
      Err(e) => {
        debug!("TRAK 0 contains no JPEG preview, is it PQ/HEIF? Error: {}", e);
        Err(RawlerError::General(
          "Unable to extract preview image from CR3 HDR-PQ file. Please see 'https://github.com/dnglab/dnglab/issues/7'".into(),
        ))
      }
    }
  }
}

impl<'a> Cr3Decoder<'a> {
  pub fn _new_makernote(buf: &'a [u8], offset: usize, base_offset: usize, chain_level: isize, e: Endian) -> Result<LegacyTiffIFD<'a>> {
    let mut off = 0;
    let data = &buf[offset..];
    let mut endian = e;

    // Some have MM or II to indicate endianness - read that
    if data[off..off + 2] == b"II"[..] {
      off += 2;
      endian = Endian::Little;
    }
    if data[off..off + 2] == b"MM"[..] {
      off += 2;
      endian = Endian::Big;
    }

    Ok(LegacyTiffIFD::new(buf, offset + off, base_offset, 0, chain_level + 1, endian, &vec![])?)
  }

  fn get_trak_index(&self, image_type: Cr3ImageType) -> Option<usize> {
    if let Some(cr3desc) = &self.bmff.filebox.moov.cr3desc {
      cr3desc.cctp.ccdts.iter().filter(|ccdt| ccdt.image_type == image_type as u64).next().map(|rec| {
        assert!(rec.trak_index > 0);
        (rec.trak_index - 1) as usize
      })
    } else {
      None
    }
  }

  fn parse_makernotes(&mut self, rawfile: &mut RawFile) -> Result<()> {
    if let Some(Entry {
      value: crate::formats::tiff::Value::Short(v),
      ..
    }) = self.cmt3.get_entry(Cr3MakernoteTag::CameraSettings)
    {
      let lens_info = v[22];
      debug!("Lens Info tag: {}", lens_info);

      if let Some(Entry {
        value: crate::formats::tiff::Value::Ascii(lens_id),
        ..
      }) = self.cmt2.get_entry(ExifTag::LensModel)
      {
        let lens_str = &lens_id.strings()[0];
        let resolver = LensResolver::new().with_lens_model(lens_str);
        self.lens_description = resolver.resolve();
      }
    }

    if let Some(Entry {
      value: crate::formats::tiff::Value::Byte(v),
      ..
    }) = self.cmt3.get_entry(Cr3MakernoteTag::ImgUniqueID)
    {
      if v.len() == 16 {
        debug!("CR3 makernote ImgUniqueID: {:x?}", v);
        self.image_unique_id = Some(v.as_slice().try_into().expect("Invalid slice size"));
      }
    }

    if let Some(entry) = self.cmt3.get_entry(0x0001) {
      let quality = match entry.get_u16(3)? {
        Some(4) => "RAW",
        Some(5) => "Superfine",
        Some(7) => "CRAW",
        Some(130) => "LightRaw",
        Some(131) => "StandardRaw",
        _ => "unknown",
      };
      debug!("Canon quality mode: {}", quality);
    }

    if let Some(entry) = self.cmt3.get_entry(0x4026) {
      let clog = match entry.get_u32(11)? {
        Some(0) => "OFF",
        Some(1) => "CLog1",
        Some(2) => "CLog2",
        Some(3) => "CLog3",
        _ => "unknown",
      };
      debug!("Canon CLog mode: {}", clog);
    }

    if let Some(xpacket_box) = self.bmff.filebox.cr3xpacket.as_ref() {
      let offset = xpacket_box.header.offset + xpacket_box.header.header_len;
      let size = xpacket_box.header.size - xpacket_box.header.header_len;
      let buf = rawfile
        .get_range(offset, size)
        .map_err(|e| RawlerError::General(format!("I/O error while reading CR3 XPACKET: {:?}", e)))?;
      self.xpacket = Some(Vec::from(buf));
    }

    Ok(())
  }

  fn load_ctmd(&mut self, rawfile: &mut RawFile, sample_idx: u32) -> Result<()> {
    if let Some(ctmd) = self.read_ctmd(rawfile, sample_idx)? {
      if let Some(rec5) = ctmd.exposure_info()? {
        debug!("CTMD Rec(5): {:?}", rec5);
        self.ctmd_exposure = Some(rec5);
      }
      self.ctmd_focallen = ctmd.focal_len()?;

      if let Some(rec7) = ctmd.get_as_tiff(7, CR3_CTMD_BLOCK_EXIFIFD)? {
        self.ctmd_rec7_exif = Some(rec7);
      }
      if let Some(rec7) = ctmd.get_as_tiff(7, CR3_CTMD_BLOCK_MAKERNOTES)? {
        self.ctmd_rec7_makernotes = Some(rec7);
      }
      if let Some(rec8) = ctmd.get_as_tiff(8, CR3_CTMD_BLOCK_MAKERNOTES)? {
        if let Some(levels) = rec8.get_entry(Cr3MakernoteTag::ColorData) {
          if let crate::formats::tiff::Value::Short(v) = &levels.value {
            if let Some(offset) = self.camera.param_usize("colordata_wbcoeffs") {
              self.wb = Some([v[offset] as f32, v[offset + 1] as f32, v[offset + 3] as f32, NAN]);
            }
            if let Some(offset) = self.camera.param_usize("colordata_blacklevel") {
              debug!("Blacklevel offset: {:x}", offset);
              self.blacklevels = Some([v[offset], v[offset + 1], v[offset + 2], v[offset + 3]]);
            }
            if let Some(offset) = self.camera.param_usize("colordata_whitelevel") {
              self.whitelevel = Some(v[offset]);
            }
          }
        }
        self.ctmd_rec8 = Some(rec8);
      }
      if let Some(rec9) = ctmd.get_as_tiff(9, CR3_CTMD_BLOCK_MAKERNOTES)? {
        self.ctmd_rec9 = Some(rec9);
      }
    }

    debug!("CR3 blacklevels: {:?}", self.blacklevels);
    debug!("CR3 whitelevel: {:?}", self.whitelevel);

    Ok(())
  }
}

/// CTMD section with multiple records
#[derive(Clone, Debug)]
struct Ctmd {
  pub records: HashMap<u16, CtmdRecord>,
}

/// Record inside CTMD section
#[allow(dead_code)]
#[derive(Clone, Debug)]
struct CtmdRecord {
  pub rec_size: u32,
  pub rec_type: u16,
  pub reserved1: u8,
  pub reserved2: u8,
  pub reserved3: u16,
  pub reserved4: u16,
  pub payload: Vec<u8>,
  pub blocks: HashMap<u16, Vec<u8>>,
}

impl Ctmd {
  pub fn new(data: &mut ByteStream) -> Self {
    let mut records = HashMap::new();

    while data.remaining_bytes() > 0 {
      let size = data.get_u32();
      let mut rec = CtmdRecord {
        rec_size: size,
        rec_type: data.get_u16(),
        reserved1: data.get_u8(),
        reserved2: data.get_u8(),
        reserved3: data.get_u16(),
        reserved4: data.get_u16(),
        payload: data.get_bytes(size as usize - 12),
        blocks: HashMap::new(),
      };
      debug!(
        "CTMD Rec {}:  {}, {}, {}, {}",
        rec.rec_type, rec.reserved1, rec.reserved2, rec.reserved3, rec.reserved4
      );
      if [7, 8, 9].contains(&rec.rec_type) {
        let mut bs = ByteStream::new(rec.payload.as_slice(), Endian::Little);
        let mut _block_id = 0;
        while bs.remaining_bytes() > 0 {
          let sz = bs.get_u32();
          let tag = bs.get_u16();
          let _uk = bs.get_u16();
          let data = bs.get_bytes(sz as usize - 8);
          //dump_buf(&format!("/tmp/ctmd_rec{}_block{}_tag0x{:X}_uk{:X}.bin", rec.rec_type, block_id, tag, uk), data.as_slice());
          if [CR3_CTMD_BLOCK_EXIFIFD, CR3_CTMD_BLOCK_MAKERNOTES].contains(&tag) {
            assert_eq!(rec.blocks.contains_key(&tag), false, "Double tag found?!");
            rec.blocks.insert(tag, data);
          }
          _block_id += 1;
        }
      } else {
        //dump_buf(&format!("/tmp/ctmd_rec{}.bin", rec.rec_type), rec.payload.as_slice());
      }
      records.insert(rec.rec_type, rec);
    }
    Self { records }
  }

  pub fn get_as_tiff(&self, record: u16, tag: u16) -> Result<Option<GenericTiffReader>> {
    if let Some(block) = self.records.get(&record).and_then(|rec| rec.blocks.get(&tag)) {
      Ok(Some(GenericTiffReader::new_with_buffer(&block, 0, 0, Some(0))?))
    } else {
      warn!("Unable to find CTMD record {}, tag 0x{:X}", record, tag);
      Ok(None)
    }
  }

  pub fn exposure_info(&self) -> Result<Option<CtmdExposureInfo>> {
    if let Some(rec) = self.records.get(&5) {
      let mut buf = ByteStream::new(rec.payload.as_slice(), Endian::Little);
      let fnumber = Rational::new(buf.get_u16().into(), buf.get_u16().into());
      let exposure = Rational::new(buf.get_u16().into(), buf.get_u16().into());
      let iso_speed = buf.get_u32();
      let unknown = buf.get_bytes(buf.remaining_bytes());
      Ok(Some(CtmdExposureInfo {
        fnumber,
        exposure,
        iso_speed,
        unknown,
      }))
    } else {
      Ok(None)
    }
  }

  pub fn focal_len(&self) -> Result<Option<Rational>> {
    if let Some(rec) = self.records.get(&4) {
      let mut buf = ByteStream::new(rec.payload.as_slice(), Endian::Little);
      let focal_len = Rational::new(buf.get_u16().into(), buf.get_u16().into());
      Ok(Some(focal_len))
    } else {
      Ok(None)
    }
  }
}

#[derive(Debug, Clone, Serialize, Deserialize, PartialEq)]
pub struct CtmdExposureInfo {
  pub fnumber: Rational,
  pub exposure: Rational,
  pub iso_speed: u32,
  pub unknown: Vec<u8>,
}

#[derive(Debug, Copy, Clone, PartialEq, enumn::N)]
#[repr(u16)]
pub enum Cr3MakernoteTag {
  CameraSettings = 0x0001,
  FocusInfo = 0x0002,
  FlashInfo = 0x0003,
  ShotInfo = 0x0004,
  Panorama = 0x0005,
  ImageType = 0x0006,
  FirmareVer = 0x0007,
  FileNumber = 0x0008,
  OwnerName = 0x0009,
  UnknownD30 = 0x000a,
  SerialNum = 0x000c,
  CameraInfo = 0x000d,
  FileLen = 0x000e,
  CustomFunc = 0x000f,
  ModelId = 0x0010,
  MovieInfo = 0x0011,
  AFInfo = 0x0012,
  ThumbArea = 0x0013,
  SerialFormat = 0x0014,
  SuperMacro = 0x001a,
  DateStampMode = 0x001c,
  MyColors = 0x001d,
  FirmwareRev = 0x001e,
  Categories = 0x0023,
  FaceDetect1 = 0x0024,
  FaceDetect2 = 0x0025,
  AFInfo2 = 0x0026,
  ContrastInfo = 0x0027,
  ImgUniqueID = 0x0028,
  WBInfo = 0x0029,
  FaceDetect3 = 0x002f,
  TimeInfo = 0x0035,
  BatteryType = 0x0038,
  AFInfo3 = 0x003c,
  RawDataOffset = 0x0081,
  OrigDecisionDataOffset = 0x0083,
  CustomFunc1D = 0x0090,
  PersFunc = 0x0091,
  PersFuncValues = 0x0092,
  FileInfo = 0x0093,
  AFPointsInFocus1D = 0x0094,
  LensModel = 0x0095,
  InternalSerial = 0x0096,
  DustRemovalData = 0x0097,
  CropInfo = 0x0098,
  CustomFunc2 = 0x0099,
  AspectInfo = 0x009a,
  ProcessingInfo = 0x00a0,
  ToneCurveTable = 0x00a1,
  SharpnessTable = 0x00a2,
  SharpnessFreqTable = 0x00a3,
  WhiteBalanceTable = 0x00a4,
  ColorBalance = 0x00a9,
  MeasuredColor = 0x00aa,
  ColorTemp = 0x00ae,
  CanonFlags = 0x00b0,
  ModifiedInfo = 0x00b1,
  TnoeCurveMatching = 0x00b2,
  WhiteBalanceMatching = 0x00b3,
  ColorSpace = 0x00b4,
  PreviewImageInfo = 0x00b6,
  VRDOffset = 0x00d0,
  SensorInfo = 0x00e0,
  ColorData = 0x4001,
  CRWParam = 0x4002,
  ColorInfo = 0x4003,
  Flavor = 0x4005,
  PictureStyleUserDef = 0x4008,
  PictureStylePC = 0x4009,
  CustomPictureStyleFileName = 0x4010,
  AFMicroAdj = 0x4013,
  VignettingCorr = 0x4015,
  VignettingCorr2 = 0x4016,
  LightningOpt = 0x4018,
  LensInfo = 0x4019,
  AmbienceInfo = 0x4020,
  MultiExp = 0x4021,
  FilterInfo = 0x4024,
  HDRInfo = 0x4025,
  AFConfig = 0x4028,
  RawBurstModeRoll = 0x403f,
}

impl Into<u16> for Cr3MakernoteTag {
  fn into(self) -> u16 {
    self as u16
  }
}

impl TryFrom<u16> for Cr3MakernoteTag {
  type Error = String;

  fn try_from(value: u16) -> std::result::Result<Self, Self::Error> {
    Self::n(value).ok_or(format!("Unable to convert tag: {}, not defined in enum", value))
  }
}

impl TiffTagEnum for Cr3MakernoteTag {}
